/********************************************************************
*     (C) 19-Mar-1999y by Hard Wisdom "LVOV Files Viewer" v1.0      *
*     (C) 26-Jul-2020y by Hard Wisdom "LVOV Files Viewer" v1.1      *
*         --- added codepage support                                *
*         --- code cleanup                                          *
*     (C) 28-Jul-2020y by Hard Wisdom "LVOV Files Viewer" v1.2      *
*         --- bugfix: '?' instead of latin characters               *
********************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

enum ErrorCodes {
    erOK = 0,
    erIO = 1,
    erLogic = 2,
    erParam = 3,
    erHardw = 4,
    erCancel = 5
};

/*-----------------------------------------------------------------*/
char* token[]={
    /* 0x80 */
    "END", "FOR", "NEXT", "DATA", "INPUT", "DIM", "READ", "LET",
    "GOTO", "RUN", "IF", "RESTORE", "GOSUB", "RETURN", "REM", "STOP",
    /* 0x90 */
    "CLS", "ON", "PLOT", "DRAW", "POKE", "PRINT", "DEF", "CONT",
    "LIST", "CLEAR", "CLOAD", "CSAVE", "NEW", "LOAD", "MERGE", "SAVE",
    /* 0xA0 */
    "OUT", "WAIT", "SLOAD", "BSAVE", "BLOAD", "BAUD", "LOCATE", "COLOR",
    "LINE", "PSET", "PRESET", "CIRCLE", "PAINT", "GET", "PUT", "BEEP",
    /* 0xB0 */
    "SOUND", "DELETE", "RENUM", "EDIT", "VPOKE", "@@", "@@", "@@",
    "@@", "@@", "@@", "@@", "TAB(", "TO", "SPC(", "FN",
    /* 0xC0 */
    "THEN", "NOT", "STEP", "+", "-", "*", "/", "^",
    "AND", "OR", ">", "=", "<", "@@", "@@", "@@",
    /* 0xD0 */
    "SGN", "INT", "ABS", "USR", "FRE", "INP", "POS", "SQR",
    "RND", "LOG", "EXP", "COS", "SIN", "TAN", "ATN", "PEEK",
    /* 0xE0 */
    "LEN", "STR$", "VAL", "ASC", "CHR$", "LEFT$", "RIGHT$", "MID$",
    "INKEY$", "HEX$", "POINT", "VARPTR", "CSRLIN", "VPEEK", "@@", "@@",
    /* 0xF0 */
    "@@", "@@", "@@", "@@", "@@", "@@", "@@", "@@",
    "@@", "@@", "@@", "@@", "@@", "@@", "@@", "@@"
};

/*-----------------------------------------------------------------*/
unsigned int CodePage = 1251;

int xcode(int i){
    if (i < 0x60) return i;
    if (i > 0x7F) return '?';

    switch (CodePage) {
    default:
    case 7: /* KOI-7 */
        return i;

    case 8: /* KOI-8 */
        return i + 0x80;

    case 866: /* OEM CP 866 */
        return
            "\x9E\x80\x81\x96\x84\x85\x94\x83"
            "\x95\x88\x89\x8A\x8B\x8C\x8D\x8E"
            "\x8F\x9F\x90\x91\x92\x93\x86\x82"
            "\x9C\x9B\x87\x98\x9D\x99\x97\x9A"
            [i - 0x60];

    case 1251: /* ANSI CP Windows 1251 */
        return
            "\xDE\xC0\xC1\xD6\xC4\xC5\xD4\xC3"
            "\xD5\xC8\xC9\xCA\xCB\xCC\xCD\xCE"
            "\xCF\xDF\xD0\xD1\xD2\xD3\xC6\xC2"
            "\xDC\xDB\xC7\xD8\xDD\xD9\xD7\xDA"
            [i - 0x60];
    }
}

/*-----------------------------------------------------------------*/
unsigned long ByteSum;

int getByte(FILE* f) {
    int b = fgetc(f);
    if (b != EOF) ByteSum += b;
    return b;
}

/*-----------------------------------------------------------------*/
void ShowFile(char* name_of_file) {
    FILE* f;
    unsigned char sign[16], line[17];
    char name_of_tape[7];

    unsigned int beg, end, run;
    unsigned long pos;
    int i;

    if ((f = fopen(name_of_file, "rb")) == NULL) {
        printf("Error: Unable to open '%s' !\n", name_of_file);
        return;
    }

    fread(sign, 16, 1, f);
    if (strncmp("LVOV/2.0/", sign, 9)) {
        fclose(f);
        printf("Error: '%s' - Unsupported file version !\n", name_of_file);
        return;
    }

    strncpy(name_of_tape, &sign[10], 6); name_of_tape[6] = 0;
    printf("File: %s / %s ", name_of_file, name_of_tape);

    switch(sign[9]) {
        case 0xD0:
            fread(sign, 6, 1, f);
            beg = ((unsigned int)sign[1] << 8) + sign[0];
            end = ((unsigned int)sign[3] << 8) + sign[2];
            run = ((unsigned int)sign[5] << 8) + sign[4];
            printf("BLOAD %.4Xh-%.4Xh,%.4Xh\n\n", beg, end, run);

            line[16] = 0;

            for (ByteSum = i = 0, pos = beg; pos <= end; pos++) {
                if (!i) printf("%.4lX:   ",pos);
                printf("%.2X%s", line[i] = getByte(f), i == 7 ? " - " : " ");
                if (line[i] < 32 || line[i] > 127) line[i] = '.';
                else line[i] = xcode(line[i]);
                if (i < 15) i++; else {
                    i = 0; printf("  *%s*\n", line);
                }
            }
            if (i) {
                while (i < 16) {
                    printf("%s", i == 7 ? "   - " : "   ");
                    line[i++] = '.';
                }
                printf("  *%s*\n", line);
            }
            printf("\nChkSum = %.lXh\n", ByteSum);
            break;

        case 0xD3: puts("CLOAD\n");
            ByteSum = 0;
            while (getByte(f) + getByte(f)) {
                pos = getByte(f) + ((unsigned long)getByte(f) << 8);
                printf("%lu ", pos);
                do {
                    i = getByte(f);
                    if (i & 0x80) printf("%s", token[i & 0x7F]);
                    else putchar(xcode(i));
                }
                while (i);
                puts("");
            }
            printf("\nChkSum = %.lXh\n", ByteSum);
            break;

        default:
            puts("- Unknown file type!");
    }
    fclose(f);
    puts("\n");
}

/*-----------------------------------------------------------------*/
/*                      E N T R Y   P O I N T                      */
/*-----------------------------------------------------------------*/
int main (int argc, char* argv[]) {
    int i;

    puts(
        "(C) 28-Jul-2020y by Hard Wisdom 'LVOV Files Viewer' v1.2\n"
        "                                ~~~~~~~~~~~~~~~~~~~"
    );
    if (argc<2) {
        puts(
            "USAGE: LV_VIEW.EXE [options] <BinaryFile>\n"
            "       where BinaryFile - is the file parsed from the tape\n"
            "       and options is one of\n"
            "           -cp=866 MS-DOS OEM code page\n"
            "           -cp=1251 Windows ANSI code page\n"
            "           -cp=koi7 LVOV KOI-7 code page\n"
            "           -cp=koi8 Unix KOI-8 code page"
        );
        return erParam;
    }
    for (i = 1; i < argc; i++) {
        if (strcmp("-cp=koi7", argv[i]) == 0) { CodePage = 7; continue; }
        if (strcmp("-cp=koi8", argv[i]) == 0) { CodePage = 8; continue; }
        if (strcmp("-cp=866" , argv[i]) == 0) { CodePage = 866; continue; }
        if (strcmp("-cp=1251", argv[i]) == 0) { CodePage = 1251; continue; }
        ShowFile(argv[i]);
    }
    return erOK;
}
